/* ***********************************************************************************************
*mecombobox.c													
*(c)Eduardo Rodriguez 2005																
*																					
*porpuse:   This creates a Custom combo box for win32, intended to be schematic enough so as to
*           provide the basic functionality and allow for easy changes in the sourcecode, thus
*           not having to rely on 3rd party libraries and keeping the code tidy and small.
*
*history:	21/09/2005 created	
TODO
The control doesn't get the keyboard focus when the tab key is pressed. There doesn't
seem to be a simple solution to this, other than creating a tab-focus manager that
would hook on the keyboard and set and kill focus as appropiate. Controls (win32 or ME)
would have to register with the manager to get focus.														
*********************************************************************************************** */


#include <windows.h>
#include "MEControls.h"

/*************************
*DATA MEMBERS
**************************/
//info for every instance
typedef struct _MECombo
{
    int width;    
    int height; 
    int dropwidth;
    int dropheight; 
    int currentSel; 
    int fontSize;
    int widthButton;
    int elements;
    BOOL over;
    BOOL pressed;
    char* fontName;        
    HWND  hwnd;
    char* string;        
}MECombo;


const char g_MECombo[]="mecombo";
extern HDC memdc;
extern HBITMAP bm;



//////////////////////////////////////////////////////////////////////////////////////
/***********************************METHODS ("PRIVATE")*****************************/
///////////////////////////////////////////////////////////////////////////////////////
//Set and retrieve window info
static MECombo*  GetMECombo(HWND hwnd)
{
    return (MECombo *)GetWindowLong(hwnd, 0);
}

static void SetMECombo(HWND hwnd, MECombo *ccp)
{
    SetWindowLong(hwnd, 0, (LONG)ccp);
}




/*****************************
*PAINT
*****************************/

static void paintMECombo(HWND hwnd,MECombo* ccp)
{
    //blablabla
    PAINTSTRUCT Ps;
    HDC hDC = BeginPaint(hwnd, &Ps);
    HDC MemDC= CreateCompatibleDC(hDC);
    int offset;
    COLORREF c1,c2;
    SelectObject(MemDC,bm); 
    c1=GetPixel(MemDC,12,92);
    c2=GetPixel(MemDC,11,92);
    SetTextColor(memdc, c2);
    SetBkColor(memdc,c1);

    //paint rectangle
    Rectangle(memdc,0,0,ccp->width,ccp->height);
    BitBlt(memdc, 0, 0,ccp->width,ccp->height , MemDC, 11, 91, SRCCOPY); 
     

    //get buffer content and write it
    char buffer[strlen(ccp->string)];
    int kk=0,ii=0,pp=0;
    while(ccp->string[ii]!='\0')
    {
        if(ccp->string[ii]==',')
            pp++;
        else
        {
            if(pp==ccp->currentSel)
            {
                buffer[kk]=ccp->string[ii];
                kk++;                     
            }               
        }                 
        ii++;
    }
    buffer[kk]='\0';
    
    //font
    HFONT hf = CreateFont(ccp->fontSize, 0, 0, 0,0, FALSE, 0, 0, 0, 0, 0, 0, 0, ccp->fontName);   
    SelectObject(memdc,hf);
    
    //and out
    TextOut(memdc, 5, (ccp->height-ccp->fontSize)/2, buffer, strlen(buffer));   
 
    //paint button rectangle
    if(!ccp->over && !ccp->pressed)
    offset=11;
    else if(ccp->pressed)
    offset=23;
    else
    offset=35;
    BitBlt(memdc, ccp->width-ccp->widthButton-2, 2,12,12 , MemDC, offset, 71, SRCCOPY); 

    //double buffering               
    BitBlt(hDC, 0, 0,ccp->width,ccp->height , memdc, 0, 0, SRCCOPY); 

    //blablabla           
    DeleteDC(hDC);
    DeleteObject(hf);
    DeleteDC(MemDC);
    EndPaint(hwnd, &Ps);                    
}



/******************************
*CALLBACK PROCEDURE
*******************************/

LRESULT CALLBACK MEComboProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    MECombo *ccp = GetMECombo(hwnd);
    switch(msg)
    {
        
        
        //-------------------------WM_NCHITTEST---------------------------------       
        case WM_NCHITTEST:
        {
            //get coordinates
            UINT uHitTest = DefWindowProc(hwnd, WM_NCHITTEST, wParam, lParam);
            int xPos = LOWORD(lParam); 
            int yPos = HIWORD(lParam);
            RECT wind; 
            GetWindowRect(hwnd,&wind);
            
            //is it over the the button?
            if( uHitTest==HTCLIENT && !ccp->over)
            { 
                if(xPos>ccp->width-ccp->widthButton-2+wind.left)
                {
                    TRACKMOUSEEVENT tr;
                    tr.cbSize=sizeof(tr);
                    tr.dwFlags=TME_LEAVE;
                    tr.hwndTrack=hwnd;
                    tr.dwHoverTime=HOVER_DEFAULT;
                    TrackMouseEvent(&tr);
                    ccp->over=TRUE;
         	        InvalidateRect(hwnd,NULL,FALSE);        
                    UpdateWindow(hwnd);                         
                } 
            }
            
            //and if not
            else if((ccp->over || ccp->pressed)&& xPos<ccp->width-ccp->widthButton-2+wind.left)
            {
                ccp->over=FALSE;
                ccp->pressed=FALSE;
                InvalidateRect(hwnd,NULL,FALSE);        
                UpdateWindow(hwnd);    
            }
            
            //blablabla
            return uHitTest;  
        }     
        break; 
        
        
        //--------------------------WM_LBUTTONUP-------------------------------------
        case WM_LBUTTONUP:
            if(ccp->pressed)
            {
                ccp->pressed=FALSE;                
                InvalidateRect(hwnd,NULL,FALSE);        
                UpdateWindow(hwnd);                            
            }
        break;
        
        
        //--------------------------WM_MOUSELEAVE-----------------------------------
        case WM_MOUSELEAVE:
            ccp->over=FALSE;
            ccp->pressed=FALSE;
            InvalidateRect(hwnd,NULL,FALSE);        
            UpdateWindow(hwnd);        
        break;
        
        
        //---------------------------WM_LBUTTONDOWN-------------------------------- 
        case WM_LBUTTONDOWN:
        {
            int xPos = LOWORD(lParam); 
            int yPos = HIWORD(lParam);
            if(xPos>ccp->width-ccp->widthButton-2)
            {
                TRACKMOUSEEVENT tr;
                RECT wind; 
                GetWindowRect(hwnd,&wind);
                tr.cbSize=sizeof(tr);
                tr.dwFlags=TME_LEAVE;
                tr.hwndTrack=hwnd;
                tr.dwHoverTime=HOVER_DEFAULT;
                TrackMouseEvent(&tr);
                ccp->pressed=TRUE;
  	            InvalidateRect(hwnd,NULL,FALSE);        
                UpdateWindow(hwnd); 
                CreateMEList(hwnd,wind.left,wind.top+ccp->height, ccp->dropwidth,ccp->dropheight,ccp->string,ccp->fontName,ccp->fontSize,ccp->currentSel,ccp->elements,TRUE);                                             
            }
    
        }        
        break;
        
        
        //-----------------------------WM_CHAR----------------------------------
        case WM_CHAR:
            //if it's return
            if(wParam=='\r')
            {
                RECT wind; 
                GetWindowRect(hwnd,&wind);
                CreateMEList(hwnd,wind.left,wind.top+ccp->height, ccp->dropwidth,ccp->dropheight,ccp->string,ccp->fontName,ccp->fontSize,ccp->currentSel,ccp->elements,TRUE);                                             
            }          
        break;
                 

        //---------------------------WM_MOUSEACTIVATE----------------------------
        case WM_MOUSEACTIVATE:
            SetFocus(hwnd);
        return MA_ACTIVATE;
    
        
        //--------------------------------ME_LIST_SELECTED----------------------------- 
        case ME_LIST_SELECTED:
            ccp->currentSel=wParam;
            InvalidateRect(hwnd,NULL,FALSE);        
            UpdateWindow(hwnd); 
        break;
        
        
        //--------------------------------WM_NCDESTROY-------------------------------
        case WM_NCDESTROY:
            free(ccp);
        break;
        
        
        //---------------------------------WM_PAINT----------------------------------
        case WM_PAINT:
            paintMECombo(hwnd,ccp);
        break;
        
        
        //blablabla
        default:
        break;
    }
    return DefWindowProc(hwnd, msg, wParam, lParam);
}



////////////////////////////////////////////////////////////////////////////////////
/****************************METHODS ("PUBLIC")************************************/
////////////////////////////////////////////////////////////////////////////////////

/*****************************
*REGISTER THE CLASS

return void;
HWND parent: any valid handle;
Should be initialized by Global
********************************/
void InitMEComboBox(HWND parent)
{
 
    WNDCLASSEX wc;

    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.style         = 0;
    wc.lpfnWndProc   = MEComboProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = sizeof( MECombo * );
    wc.hInstance     = GetModuleHandle(0);
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = g_MECombo;
    wc.hIconSm       = LoadIcon(NULL, IDI_APPLICATION);

    RegisterClassEx(&wc); 
}


/**************************
*CREATE COMBO BOX

return: handle to the scrollbar created;
see mecombobox.h  or MEControls.h
for an explanation of the paramaters
***************************/

HWND CreateMEComboBox(HWND parent,int posX,int posY, int w,int h,int dropWidth,int dropHeight,int wb, char* strings,char* fontName,int fontSize,int elements)
{
     //----------------------WARNING!!!!!!!!!!
     //last argument takes a single string, with elements separated by \n

 
   HWND thisone=CreateWindowEx(0,g_MECombo, "",WS_VISIBLE|WS_CHILD|WS_TABSTOP,
       posX, posY, w, h,parent, NULL, GetModuleHandle(0), NULL);  
   MECombo *ccp= GetMECombo(thisone);
   ccp = malloc(sizeof(MECombo));
   if(ccp == NULL) return NULL;
   
   // Initialize the CustCtrl structure. 
   ccp->width=w;
   ccp->height=h;
   ccp->string=strings;
   ccp->dropwidth=dropWidth;
   ccp->dropheight=dropHeight;
   ccp->currentSel=0;
   ccp->widthButton=wb;
   ccp->fontName=fontName;
   ccp->fontSize=fontSize;
   ccp->over=FALSE;
   ccp->pressed=FALSE;
   ccp->elements=elements;
   SetMECombo(thisone, ccp);
    
return thisone;  
}


int GetSelectedMEComboBox(HWND hwnd)
{
    MECombo *ccp = GetMECombo(hwnd);
    return ccp->currentSel;    
}
