/* ***********************************************************************************************
*melist.c													
*(c)Eduardo Rodriguez 2005																
*																					
*porpuse:   This creates a Custom list for win32, intended to be schematic enough so as to
*           provide the basic functionality and allow for easy changes in the sourcecode, thus
*           not having to rely on 3rd party libraries and keeping the code tidy and small.
*           there's a limit of 1000 chars for every element in the list
*history:	27/09/2005 created
TODO
Maybe: at the moment the window get's destroyed instead of hidden if the selfdestroy
parameter is set. That means it needs to be created from scratch every time it's shown,
but it is quite fast, doesn't carry much overhead apparently. So, maybe modify it so that
it hides rather than destroy.										
*********************************************************************************************** */


#include <windows.h>
#include "MEControls.h"


/*************************
*DATA MEMBERS
**************************/

//info for every instance
typedef struct _MEList
{
    int width;             
    int height;  
    int fontSize;          
    int index;             //current top element in list
    int elements;          //number of elements
    int selectElement;     //element over wich the mouse is
    char* fontName;       
    HWND hwnd;             //handle to parent
    HWND scroll;           //handle to scrollbar
    char* string;          //pointer to comma separated & terminated string containing elements. 
    BOOL selfdestroy;      //should the window self destroy when it losses focus?
}MEList;


const char g_MEList[]="melist";

extern HDC memdc;
extern HBITMAP bm;


/////////////////////////////////////////////////////////////////////////////////////////
/************************************METHODS ("PRIVATE")********************************/
/////////////////////////////////////////////////////////////////////////////////////////
//set and retrieve window info
static MEList*  GetMEList(HWND hwnd)
{
    return (MEList *)GetWindowLong(hwnd, 0);
}


static void SetMEList(HWND hwnd, MEList *ccp)
{
    SetWindowLong(hwnd, 0, (LONG)ccp);
}



/*************************
*PAINT
**************************/
//very simple painting proc, to be custumized. It does some
//calculations one shouldn't mess with, though


static void MEListPaint(HWND hwnd, MEList *ccp)
{
    //make sure it has been initiliazed 
    if(ccp!=NULL)
    {
         //blabla        
         PAINTSTRUCT Ps;
         COLORREF c1,c2,c3;
         HBRUSH bk1,bk2;
         HPEN p1,p2;
         HFONT hf = CreateFont(ccp->fontSize, 0, 0, 0,0, FALSE, 0, 0, 0, 0, 0, 0, 0, ccp->fontName);   
         int kk=0,ll=0,jj=0,nn=0,pp=1; //some counters
         char buffer[1000];  //char buffer to hold an element of the list at a time
         
         //initialize the buffer
         memset(buffer,'\0',1000);
             
         //blablabla
         HDC hDC = BeginPaint(hwnd, &Ps);
         HDC MemDC= CreateCompatibleDC(hDC);
         SelectObject(MemDC,bm);
         c1=GetPixel(MemDC,1,0);
         c2=GetPixel(MemDC,2,0);
         c3=GetPixel(MemDC,3,0);         
         SelectObject(memdc,hf);
         
         //paint bg
         bk1=CreateSolidBrush(c1);
         p1=CreatePen(PS_SOLID, 1, c2); 
         bk2=CreateSolidBrush(c3);
         p2=CreatePen(PS_SOLID,1,c3);
         SelectObject(memdc,bk1);
         SelectObject(memdc,p1);
         Rectangle(memdc,0,0,ccp->width,ccp->height);
         
         //loop through the string
         kk=0;
         SetTextColor(memdc, c2);
         while(ccp->string[kk]!='\0')
         {
           //store current element of the list                          
           if(ccp->string[kk]!=',')
           {
              buffer[ll]=ccp->string[kk];
              ll++;                   
           }
           
           //if we already finished retrieving the next element...
           else
           {
              //if the element is further than the index 
              if(jj>=ccp->index)
              {
                 SetBkColor(memdc, c1);             
                 //if this is the one the mouse is on top of...
                 if(ccp->selectElement==jj && pp<ccp->height-ccp->fontSize)
                 {
                     SetBkColor(memdc, c3);
                     SelectObject(memdc,bk2);
                     SelectObject(memdc,p2);                                
                     Rectangle(memdc,1,pp,ccp->width-1,pp+ccp->fontSize);
                 }
                 
                 
                 
                 //and if theres room for it...
                 if(pp<ccp->height-ccp->fontSize)
                     TextOut(memdc,4,pp,buffer,strlen(buffer));             
                 pp+=ccp->fontSize;                                 
              }
              //empty buffer
              ll=0;
              nn=0;
              while(buffer[nn]!='\0')
              {
                 buffer[nn]='\0';
                 nn++;
              } 
              jj++;
           }
           kk++;
         }
         
         //double buffer
         BitBlt(hDC, 0, 0,ccp->width,ccp->height , memdc, 0, 0, SRCCOPY);
         
         //blabla 
         DeleteObject(bk1);
         DeleteObject(bk2);
         DeleteObject(p1);
         DeleteObject(p2);
         DeleteDC(MemDC);
         DeleteDC(hDC);
         DeleteObject(hf);
         EndPaint(hwnd, &Ps);                   
    }     
}






/******************************
*CALLBACK PROCEDURE
*******************************/

LRESULT CALLBACK melistproc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    MEList *ccp = GetMEList(hwnd);
    switch(msg)
    {
     
        //-------------------------WM_NCHITTEST--------------------------------       
        case WM_NCHITTEST:
        {
             //check who is the mouse over and paint if necessary
             UINT uHitTest = DefWindowProc(hwnd, WM_NCHITTEST, wParam, lParam);
             int yPos = HIWORD(lParam);
             RECT wind; 
             GetWindowRect(hwnd,&wind);
             if( uHitTest==HTCLIENT)
             { 
                //compute element 
                int formerElement=ccp->selectElement;
                ccp->selectElement=ccp->index+((yPos-wind.top)/ccp->fontSize);
                if(ccp->height-(ccp->selectElement-ccp->index)*ccp->fontSize<ccp->fontSize-2)
                ccp->selectElement--;
                if(formerElement!=ccp->selectElement)
                {
                	InvalidateRect(hwnd,NULL,FALSE);        
                    UpdateWindow(hwnd);                                                        
                }
             }
             return uHitTest;
        }      
        break; 
        
        
        //--------------------------WM_NCAVTIVATE----------------------------------
        //if the message doesn't come from the scrollbar and if it is selfdestructive, destroy
        case WM_NCACTIVATE:
            if(wParam==FALSE && (HWND)lParam!=ccp->scroll && ccp->selfdestroy)
                DestroyWindow(hwnd); 
            SetFocus(hwnd);
        break; 
        
        
        //------------------------WM_LBUTTONDOWN-------------------------------------
         case WM_LBUTTONDOWN:
             SendMessage(ccp->hwnd,ME_LIST_SELECTED,ccp->selectElement,0);
             if(ccp->selfdestroy)
                 DestroyWindow(hwnd);
         break;
         
         
         //----------------------WM_MOUSEWHEEL-------------------------------------
         case WM_MOUSEWHEEL:
              ccp->index+=(short)HIWORD(-wParam)/120;
              if(ccp->index<0)
                  ccp->index=0;
              else if(ccp->index>ccp->elements-ccp->height/ccp->fontSize)
                  ccp->index=ccp->elements-ccp->height/ccp->fontSize;
              setposMEScroll(ccp->scroll,ccp->index);
 	          InvalidateRect(hwnd,NULL,FALSE);        
              UpdateWindow(hwnd);                   
         return 0;
         
         
         //------------------------ME_SCROLL_VERTICAL-------------------------------
         case ME_SCROLL_VERTICAL:
             ccp->index=wParam;
 	         InvalidateRect(hwnd,NULL,FALSE);        
             UpdateWindow(hwnd);           
         break;
    
        
         //-------------------------WM_NCDESTROY------------------------------------
         case WM_NCDESTROY:
             free(ccp);
         break;
         
         
         //---------------------------WM_PAINT-------------------------------------
         case WM_PAINT:
             MEListPaint(hwnd,ccp);
         break;
         
         
         //----------------------------WM_KEYDOWN-----------------------------------
         case WM_KEYDOWN:
              
           //down arrow
           if(wParam== VK_DOWN)
           {      
              ccp->selectElement++;
              if(ccp->selectElement>ccp->elements-1)
                  ccp->selectElement=ccp->elements-1;
              if(ccp->selectElement>=ccp->index+ccp->height/ccp->fontSize)
              {
                  ccp->index++;
              }
              if(ccp->selectElement<ccp->index)
              {
                  ccp->index=ccp->selectElement;                                 
              }
              else if(ccp->selectElement>ccp->index+(ccp->height/ccp->fontSize-1))
              {
                  ccp->index=ccp->selectElement-(ccp->height/ccp->fontSize-1);     
              }
              setposMEScroll(ccp->scroll,ccp->index);
 	          InvalidateRect(hwnd,NULL,FALSE);        
              UpdateWindow(hwnd);           
           }
           
           //up arrow
           else if(wParam==VK_UP )
           {
              ccp->selectElement--;
              if(ccp->selectElement<0)
                  ccp->selectElement=0;
              if(ccp->selectElement<ccp->index)
              {
                  ccp->index--;
              }
              if(ccp->selectElement<ccp->index)
              {
                  ccp->index=ccp->selectElement;                                 
              }
              else if(ccp->selectElement>ccp->index+(ccp->height/ccp->fontSize-1))
              {
                   ccp->index=ccp->selectElement-(ccp->height/ccp->fontSize-1);     
              }
              setposMEScroll(ccp->scroll,ccp->index);
 	          InvalidateRect(hwnd,NULL,FALSE);        
              UpdateWindow(hwnd);                   
           }
           
           //left or right arrow
           else if(wParam==VK_LEFT || wParam==VK_RIGHT)
           {
               SendMessage(ccp->hwnd,ME_LIST_SELECTED,ccp->selectElement,0);
               if(ccp->selfdestroy)
                   DestroyWindow(hwnd);                                        
           }
           break;
         
         
         //-----------------------------WM_CHAR----------------------------------
         case WM_CHAR:
             //if it's return
             if(wParam=='\r')
             {
                 SendMessage(ccp->hwnd,ME_LIST_SELECTED,ccp->selectElement,0);
                 if(ccp->selfdestroy)
                     DestroyWindow(hwnd);                        
             }          
         break;
         
         
         //blabla    
         default:
         break;
    }
    return DefWindowProc(hwnd, msg, wParam, lParam);
}



///////////////////////////////////////////////////////////////////////////////////////
/******************************METHODS ("PUBLIC")************************************/
//////////////////////////////////////////////////////////////////////////////////////

/*****************************
*REGISTER THE CLASS

return void;
HWND parent: any valid handle;
Should be initialized by Global
********************************/
void InitMEList(HWND parent)
{ 
    WNDCLASSEX wc;

    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.style         =  0;
    wc.lpfnWndProc   = melistproc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = sizeof( MEList * );
    wc.hInstance     = GetModuleHandle(0);
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = g_MEList;
    wc.hIconSm       = LoadIcon(NULL, IDI_APPLICATION);
    RegisterClassEx(&wc); 
}






/*****************************
*CREATE THE WINDOW

return handle to the newly created list;
See melist.h or MEControls.h for and explanation
of the parameters.
********************************/
HWND CreateMEList(HWND parent,int posX,int posY, int w,int h, char* strings,char* fontName,int fontSize,int index,int elements, BOOL selfdestroy)
{
     //----------------------WARNING!!!!!!!!!!
     //string argument takes a single string, with elements separated and terminated by a comma



    HWND thisone=CreateWindowEx(0,g_MEList, "",WS_POPUP|WS_CLIPCHILDREN ,posX, posY, 
        w, h,parent, NULL, GetModuleHandle(0), NULL);  
    ShowWindow (thisone,  SW_SHOW); 
    MEList *ccp = GetMEList(thisone);
    ccp = malloc( sizeof(MEList) );
    if(ccp == NULL) return NULL;
    
    // Initialize the CustCtrl structure. 
    ccp->height=h;
    ccp->width=w;
    ccp->fontName=fontName;
    ccp->fontSize=fontSize;
    ccp->string=strings;
    ccp->selectElement=index;
    if(index<elements-h/fontSize)
      ccp->index=index;
    else
      ccp->index=elements-h/fontSize;
    ccp->hwnd=parent;
    ccp->elements=elements;
    ccp->selfdestroy=selfdestroy;
     SetMEList(thisone,ccp);
         
    //if necessary, get a scrollbar
    if(ccp->elements*ccp->fontSize>ccp->height)    
       ccp->scroll=CreateMEScroll(thisone,ccp->width-12,1,11,ccp->height-2,elements-1,ccp->index,4,TRUE);    
    

    return thisone;
     
}


