/* ***********************************************************************************************
*mecombobox.c													
*(c)Eduardo Rodriguez 2005																
*																					
*porpuse:   This creates a Custom scrollbar for win32, intended to be schematic enough so as to
*           provide the basic functionality and allow for easy changes in the sourcecode, thus
*           not having to rely on 3rd party libraries and keeping the code tidy and small.
*
*history:	14/10/2005 created
TODO
2.First click should not call animateScroll, just call timer and wait for
   its answer to call animate, if the button is down

															
*********************************************************************************************** */


#include <windows.h>
#include "MEControls.h"

/*************************
*DATA MEMBERS
**************************/

//Define the posible states of the window

#define ME_SCROLL_INACTIVE -1
#define ME_SCROLL_BUTTON1 1
#define ME_SCROLL_BUTTON2 2
#define ME_SCROLL_HANDLE 3
#define ME_SCROLL_SPC1 4
#define ME_SCROLL_SPC2 5




//info for every instance
typedef struct _MEScroll
{
    int width;
    int height;
    int posX;
    int posY;
    int screenSize;           //visible length. Necessary to calculate size of handle
    unsigned int docLength;   //document length in document units 
    unsigned int docPos;      //document position in document units
    int handlePos;            //handle position in pixels, not counting first button
    int handleLength;         //handle length in pixels
    BOOL vertical;            //is it a vertical scrollbar?
    BOOL disabled;            //is it disabled? find out at painting time
	HWND hwnd;				 //handle to parent
	int clickedon;           //holds position of click when clicked on handle or spaces
	int state;               //state of the window
}MEScroll;


const char g_MEScroll[]="mescrollbar";
extern HBITMAP bm;
extern HDC memdc;

///////////////////////////////////////////////////////////////////////////////////////
/**********************************METHODS ("PRIVATE")*******************************/
////////////////////////////////////////////////////////////////////////////////////////
//set and retrieve window info
static MEScroll*  GetMEScroll(HWND hwnd)
{
    return (MEScroll *)GetWindowLong(hwnd, 0);
}

static void SetMEScroll(HWND hwnd, MEScroll *ccp)
{
    SetWindowLong(hwnd, 0, (LONG)ccp);
}

/*****************************
*PAINT
*****************************/
//very simple painting proc, to be custumized. It does some
//calculations one shouldn't mess with, though


static void paintMEScroll(HWND hwnd,MEScroll* ccp)
{
     //blablabla
     PAINTSTRUCT Ps;
     HDC hDC = BeginPaint(hwnd, &Ps);
     HDC MemDC= CreateCompatibleDC(hDC);
     int kk;
     HBRUSH bk1=NULL;
     HPEN bk2=NULL;
     SelectObject(MemDC,bm);
     //paint background
     bk1=CreateSolidBrush(GetPixel(MemDC,0,0));
     bk2=CreatePen(PS_SOLID, 1, GetPixel(MemDC,0,0)); 
     SelectObject(memdc,bk1);
     SelectObject(memdc,bk2);
     Rectangle(memdc,0,0,ccp->width,ccp->height);
     

     if(!ccp->disabled)
     {
                   
         
     //--------------------VERTICAL
               if(ccp->vertical && !ccp->disabled)
               {  
                 //CALCULATIONS do not edit              
                 //handle length
                 ccp->handleLength=((ccp->height-ccp->width*2)*ccp->screenSize)/(ccp->docLength);
                 if(ccp->handleLength<8)ccp->handleLength=8;
                 //handle Pos
                 ccp->handlePos=((ccp->height-ccp->width*2)*(ccp->docPos))/(ccp->docLength);
                 //check if we've gone too far
                 if(ccp->handlePos>(ccp->height-ccp->width*2)-ccp->handleLength)
                    ccp->handlePos=(ccp->height-ccp->width*2)-ccp->handleLength;
                 //correct size to make up for decimal innacuracies
                 if(ccp->docPos==(ccp->docLength-ccp->screenSize)+1)
                    ccp->handleLength=(ccp->height-ccp->width*2)-ccp->handlePos;
                    
                    
                 //-------------PAINT
                 
                 if(ccp->height>ccp->width*2)
                 {
                     //button 1
                     if(ccp->state==ME_SCROLL_BUTTON1)   
                     BitBlt(memdc, 0, 0,ccp->width,ccp->width , MemDC, 0, 32, SRCCOPY); 
                     else
                     BitBlt(memdc, 0, 0,ccp->width,ccp->width , MemDC, 0, 21, SRCCOPY);  
                 
                     //button 2 
                     if(ccp->state==ME_SCROLL_BUTTON2)   
                     BitBlt(memdc, 0, ccp->height-ccp->width,ccp->width,ccp->width , MemDC, 0, 65, SRCCOPY); 
                     else
                     BitBlt(memdc, 0, ccp->height-ccp->width,ccp->width,ccp->width , MemDC, 0, 54, SRCCOPY); 
                 }
                 
                 //handle
                 if(ccp->height>ccp->width*2+ccp->handleLength)
                 {
                     if(ccp->state!=ME_SCROLL_HANDLE)
                     {
                         for(kk=ccp->handlePos+ccp->width;kk<ccp->handlePos+ccp->width+ccp->handleLength;kk++)
                             BitBlt(memdc, 0, kk,ccp->width,1 , MemDC, 0, 19, SRCCOPY);
                         BitBlt(memdc, 0, ccp->handlePos+ccp->width,ccp->width,4 , MemDC, 0, 1, SRCCOPY);
                         BitBlt(memdc, 0, ccp->handlePos+ccp->width+ccp->handleLength-4,ccp->width,4 , MemDC, 0, 6, SRCCOPY);                         
                         
                     }
                     else
                     {
                         for(kk=ccp->handlePos+ccp->width;kk<ccp->handlePos+ccp->width+ccp->handleLength;kk++)
                             BitBlt(memdc, 0, kk,ccp->width,1 , MemDC, 0, 20, SRCCOPY); 
                         BitBlt(memdc, 0, ccp->handlePos+ccp->width,ccp->width,4 , MemDC, 0, 10, SRCCOPY);
                         BitBlt(memdc, 0, ccp->handlePos+ccp->width+ccp->handleLength-4,ccp->width,4 , MemDC, 0, 15, SRCCOPY);                         
                                        
                     }            
                 }
        
           }
               
               
               
     //-------------------HORIZONTAL
               else if(!ccp->disabled)
               {
                 //CALCULATIONS do not edit
                 //handle length
                 ccp->handleLength=((ccp->width-ccp->height*2)*ccp->screenSize)/(ccp->docLength);
                 if(ccp->handleLength<8)ccp->handleLength=8;
                 //handle Pos
                 ccp->handlePos=((ccp->width-ccp->height*2)*(ccp->docPos))/(ccp->docLength);
                 if(ccp->handlePos>ccp->width-(ccp->height*2+ccp->handleLength))
                    ccp->handlePos=ccp->width-(ccp->height*2+ccp->handleLength);
                    
                 //PAINT
                 Rectangle(hDC,0,0,ccp->height,ccp->height);                 
                 Rectangle(hDC,ccp->width-ccp->height,0,ccp->width,ccp->height); 
                 Rectangle(hDC,ccp->handlePos+ccp->height,0,ccp->handlePos+ccp->height+ccp->handleLength,ccp->height);       
               }
               
     //double buffering
     BitBlt(hDC, 0, 0,ccp->width,ccp->height , memdc, 0, 0, SRCCOPY); 
     
     } 
        
     //blablabla
     DeleteObject(bk1);
     DeleteObject(bk2);
     DeleteDC(hDC);
     DeleteDC(MemDC);
     EndPaint(hwnd, &Ps);                       
}







/**************************
*ANIMATE THE SCROLLING
***************************/


static void animateMEScrollBar(HWND hwndMain,MEScroll* ccp)
{
        MSG messages;
        int count=0;        //count the loop, to know when to accelerate 
        int aproxperpixel=0;//optimization, aproximated doc units to scrollbar pixels
        int acc=0;          //how many doc units we've done.This two to avoid repainting unnecessarily
        int what=0;         //what message do we send?
        
        
        //if vertical
        if(ccp->vertical)
        {
            aproxperpixel=ccp->docLength/(ccp->height-ccp->width*2);
            what=ME_SCROLL_VERTICAL;    
        }
        
        //if horizontal
        else
        {
            aproxperpixel=ccp->docLength/(ccp->width-ccp->height*2);
            what=ME_SCROLL_HORIZONTAL;
        }
        
        
   //BEGING LOOP     
   while(ccp->state!=ME_SCROLL_INACTIVE)
   {        
        //repaint, just in case
        InvalidateRect(hwndMain,NULL,FALSE);        
        UpdateWindow(hwndMain);  

        //first off, check for messages   
        if(PeekMessage (&messages, NULL, 0, 0,PM_REMOVE))
        {
            TranslateMessage(&messages);
            DispatchMessage(&messages);
        }   
        
        //then, take actions depending on the state   

		//button1
        if(ccp->state==ME_SCROLL_BUTTON1 && ccp->docPos>0)
        {
            ccp->docPos--;
            acc--;                                 
        }
		//button2
        else if(ccp->state==ME_SCROLL_BUTTON2 && ccp->docPos<=ccp->docLength-ccp->screenSize)
        {
            ccp->docPos++; 
            acc++;    
        }
		//space1
        else if(ccp->state==ME_SCROLL_SPC1)
        {
            if(ccp->docPos>=ccp->clickedon+ccp->screenSize)
            ccp->docPos-=ccp->screenSize;
            else
            {
              ccp->docPos=ccp->clickedon;
              ccp->state=ME_SCROLL_INACTIVE;
            }
            acc-=ccp->screenSize;
        }
		//space2
        else if(ccp->state==ME_SCROLL_SPC2)
        {
            ccp->docPos+=ccp->screenSize;
            acc+=ccp->screenSize;
            if(ccp->docPos>=ccp->clickedon)
            {
              ccp->docPos=ccp->clickedon;
              ccp->state=ME_SCROLL_INACTIVE;
            }             
        }
        
        //shall we paint?
        if(acc>aproxperpixel || acc<0-aproxperpixel)
        {
            InvalidateRect(hwndMain,NULL,FALSE);        
            UpdateWindow(hwndMain);
            acc=0;
        }          
        
       //send message 
       SendMessage(ccp->hwnd,what,ccp->docPos,0); 
       
       //how long should we sleep?      
       if(count==0) 
       {
         Sleep(150);
         count++;
       } 
       else 
       Sleep(20);
       if(PeekMessage (&messages, NULL, 0, 0,PM_REMOVE))
        {
            TranslateMessage(&messages);
            DispatchMessage(&messages);
        }          
        
             
   }
   //repaint, just in case
   InvalidateRect(hwndMain,NULL,FALSE);        
   UpdateWindow(hwndMain);  
}










/******************************
*CALLBACK PROCEDURE
*******************************/

LRESULT CALLBACK MEScrollProc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
  MEScroll *ccp = GetMEScroll(hwnd);
  switch(msg)
    {
  
  
    //---------------------------WM_NCHITTEST-------------------------           
    case WM_NCHITTEST:
    {
      //find coordinates   
      int xPos = LOWORD(lParam); 
      int yPos = HIWORD(lParam);
      int pos,w,h;
      RECT wind;
      GetWindowRect(hwnd,&wind);
      
      //adjust them depending on wether we are vertical or not.
      if(ccp->vertical)
      {
         pos=yPos-wind.top;
         w=ccp->width;
         h=ccp->height;
      }
      else
      {
         pos=xPos-wind.left;
         w=ccp->height;    
         h=ccp->width;
      }
      
      
      
      //if there's something going on
      if(ccp->state!=ME_SCROLL_INACTIVE)
      {
            
         //check the mouse didnt leave a button or space while pressing           
         if(ccp->state==ME_SCROLL_BUTTON1 && pos>w)
            ccp->state=ME_SCROLL_INACTIVE;
         else if(ccp->state==ME_SCROLL_BUTTON2 && pos<h-w)
            ccp->state=ME_SCROLL_INACTIVE;
         else if(ccp->state==ME_SCROLL_SPC1 && pos<=w)
            ccp->state=ME_SCROLL_INACTIVE;
         else if(ccp->state==ME_SCROLL_SPC2 && pos>=h-w)
            ccp->state=ME_SCROLL_INACTIVE;
            
         //adjust the clickedon variable in case the mouse moved while pressing space   
         if(ccp->state==ME_SCROLL_SPC1)
         {
            int clicked=pos-w;
            ccp->clickedon=clicked*ccp->docLength/(h-w*2);                      
         }             
         else if(ccp->state==ME_SCROLL_SPC2)
         {
            int clicked=(pos-w)-ccp->handleLength;
            ccp->clickedon=clicked*ccp->docLength/(h-w*2);              
         }
         
         
         
      }
    }     
    return  DefWindowProc(hwnd, msg, wParam, lParam);
    
    
    //--------------------------WM_LBUTTONUP------------------------------
    case WM_LBUTTONUP:
    {
      if(ccp->state==ME_SCROLL_HANDLE)
        ReleaseCapture();
        ccp->state=ME_SCROLL_INACTIVE;
        InvalidateRect(hwnd,NULL,FALSE);        
        UpdateWindow(hwnd);        
    }
    break;
    
    
    //----------------------------WM_MOUSELEAVE----------------------------
    case WM_MOUSELEAVE:
        if(ccp->state!=ME_SCROLL_HANDLE)
           ccp->state=ME_SCROLL_INACTIVE;
    break; 
    
    
    //-----------------------------WM_LBUTTONDOWN----------------------
    case WM_LBUTTONDOWN:
    {
     //if not disabled
     if(!ccp->disabled)
     {     
         //find coordinates  
         int xPos = LOWORD(lParam); 
         int yPos = HIWORD(lParam);
         int pos,w,h;
         
         //adjust to vertical or horizontal
         if(ccp->vertical)
         {
             pos=yPos;
             w=ccp->width;
             h=ccp->height;                 
         }
         else
         {
             pos=xPos;
             w=ccp->height;    
             h=ccp->width;
         }
         
         //first button
         if(pos<w)
         {
            ccp->state=ME_SCROLL_BUTTON1;         
         } 
         
         //second button
         else if(pos>h-w)
         {
            ccp->state=ME_SCROLL_BUTTON2;    
         } 
         
         //firts space
         else if(pos>w && pos<ccp->handlePos+w)
         {
            int clicked=pos-w;
            ccp->clickedon=clicked*ccp->docLength/(h-w*2);
            ccp->state=ME_SCROLL_SPC1;    
         }  
         
         //second space
         else if(pos>ccp->handlePos+ccp->handleLength+w && pos<h-w)
         {
            int clicked=(pos-w)-ccp->handleLength;
            //translate to doc units
            ccp->clickedon=clicked*ccp->docLength/(h-w*2);
            ccp->state=ME_SCROLL_SPC2;   
         }
         
         //default: handle
         else
         {
            ccp->state=ME_SCROLL_HANDLE;  
            ccp->clickedon=pos-(ccp->handlePos+w);
            InvalidateRect(hwnd,NULL,FALSE);        
            UpdateWindow(hwnd);  
            SetCapture(hwnd);
         }
         
         //if it's not handle and not inactive, start painting and track mouse
         if(ccp->state!=ME_SCROLL_INACTIVE && ccp->state!=ME_SCROLL_HANDLE)
         { 
            TRACKMOUSEEVENT tr;
            tr.cbSize=sizeof(tr);
            tr.dwFlags=TME_LEAVE;
            tr.hwndTrack=hwnd;
            tr.dwHoverTime=HOVER_DEFAULT;
            TrackMouseEvent(&tr);    
            animateMEScrollBar(hwnd,ccp);        
         }  
      

     }
           
    }    
    break; 
    
    //------------------------------WM_MOUSEMOVE-----------------------------
    case WM_MOUSEMOVE:
    {
         //get coordinates
         int xPos = LOWORD(lParam); 
         int yPos = HIWORD(lParam);
         //what tipe of message do we send?
         int what=0;
         if(ccp->vertical)
             what=ME_SCROLL_VERTICAL;
         else
             what=ME_SCROLL_HORIZONTAL;
             
         //if VERTICAL
         if(ccp->state==ME_SCROLL_HANDLE && ccp->vertical)
         {
             //where are we in the scrollbar?                         
             int clicked=(yPos-ccp->width)-ccp->clickedon;
             if(clicked<0 || clicked>3000)
                 clicked=0;
             //and in the document?
             ccp->docPos=clicked*ccp->docLength/(ccp->height-ccp->width*2);  
             //if we are too far down,correct
             if(ccp->docPos>(ccp->docLength-ccp->screenSize)+1)ccp->docPos=(ccp->docLength-ccp->screenSize)+1;               
             //and send message and repaint
             SendMessage(ccp->hwnd,what,ccp->docPos,0);      
             InvalidateRect(hwnd,NULL,FALSE);        
             UpdateWindow(hwnd);          
         } 
         
         //if HORIZONTAL (redundant)  
         else if(ccp->state==ME_SCROLL_HANDLE && !ccp->vertical)
         {
             //where are we in the scrollbar?
             int clicked=(xPos-ccp->height)-ccp->clickedon;
             if(clicked<0 || clicked>3000)
                 clicked=0;
             //and in the document?
             ccp->docPos=clicked*ccp->docLength/(ccp->width-ccp->height*2); 
             //if we are too far down, correct 
             if(ccp->docPos>(ccp->docLength-ccp->screenSize)+1)
                 ccp->docPos=(ccp->docLength-ccp->screenSize)+1;    
             //and send message and paint       
             SendMessage(ccp->hwnd,what,ccp->docPos,0);      
             InvalidateRect(hwnd,NULL,FALSE);        
             UpdateWindow(hwnd);                 
         }   
               
    }   
    break;
    
    
    //--------------------------------WM_NCDESTROY--------------------------
    case WM_NCDESTROY:
        free(ccp);
    break;
    
    
    //--------------------------------WM_PAINT-----------------------------
    case WM_PAINT:
    {
        paintMEScroll(hwnd,ccp);                
    }
    break;
    
        //--------------------------WM_ERASEBKGND-----------------------------------
        case WM_ERASEBKGND:
        return 1;
    
    
    
    //blablabla
    default:
        break;
  }

  return DefWindowProc(hwnd, msg, wParam, lParam);
}



////////////////////////////////////////////////////////////////////////////////////////
/*******************************METHODS ("PUBLIC")***********************************/
///////////////////////////////////////////////////////////////////////////////////////

/*****************************
*REGISTER THE CLASS

return void;
HWND parent: any valid handle;
Should be initialized by Global
********************************/
void InitMEScroll(HWND parent)
{
    WNDCLASSEX wc;
    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.style         = 0;
    wc.lpfnWndProc   = MEScrollProc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = sizeof( MEScroll * );
    wc.hInstance     = GetModuleHandle(0);
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = g_MEScroll;
    wc.hIconSm       = LoadIcon(NULL, IDI_APPLICATION);
    RegisterClassEx(&wc); 
}






/**************************
*CREATE SCROLLBAR

return: handle to the scrollbar created;
see mescrollbar.h  or MEControls.h
for an explanation of the paramaters
***************************/
HWND CreateMEScroll(HWND parent,int posX,int posY, int w,int h,unsigned int docLen,unsigned int docPos,int screenSize,BOOL vertical)
{
   
 
   HWND thisone=CreateWindowEx(0,g_MEScroll, "",WS_CHILD,
      posX, posY, w, h,parent, NULL, GetModuleHandle(0), NULL);  
   MEScroll *ccp = GetMEScroll(thisone);
   ccp = malloc( sizeof(MEScroll) );
   if(ccp == NULL) return NULL;
        // Initialize the MEScroll structure. 
   ccp->width=w;
   ccp->height=h;
   ccp->hwnd=parent;
   ccp->posX=posX;
   ccp->posY=posY;
   ccp->docLength=docLen;
   ccp->docPos=docPos;
   ccp->screenSize=screenSize;
   ccp->vertical=vertical;
   ccp->state=ME_SCROLL_INACTIVE;
        SetMEScroll(thisone, ccp);
    if(ccp->docLength>ccp->screenSize)
    {
        ccp->disabled=FALSE;
        ShowWindow(thisone,SW_SHOW);
        InvalidateRect(thisone,NULL,FALSE);        
        UpdateWindow(thisone); 
    }
    else 
    {
        ccp->disabled=TRUE;
        ShowWindow(thisone,SW_HIDE);        
    }    
   return thisone;   
}










/**************************
*RESIZE

return: void
see mescrollbar.h  or MEControls.h
for an explanation of the paramaters
***************************/
void resizeMEScroll(HWND hwnd,int x,int y)
{
  if(hwnd!=NULL)
  {
    MEScroll *ccp = GetMEScroll(hwnd);
    ccp->width=x;
    ccp->height=y; 
    SetWindowPos(hwnd,0,ccp->posX,ccp->posY,x,y,0);   
    InvalidateRect(hwnd,NULL,FALSE);        
    UpdateWindow(hwnd); 
  }  
}






/**************************
*SET the POSITION within the document where we are

return: void
see mescrollbar.h  or MEControls.h
for an explanation of the paramaters
***************************/
void setposMEScroll(HWND hwnd,int pos)
{
  if(hwnd!=NULL)
  {
    MEScroll *ccp = GetMEScroll(hwnd);
    if(pos<(ccp->docLength-ccp->screenSize)+1)
    ccp->docPos=pos;
    else
    ccp->docPos=(ccp->docLength-ccp->screenSize)+1;
    InvalidateRect(hwnd,NULL,FALSE);        
    UpdateWindow(hwnd);                               
  }     
}






/**************************
*SET the DDCUMENT'S LENGTH in document units

return: void
see mescrollbar.h  or MEControls.h
for an explanation of the paramaters
***************************/
void setdoclengthMEScroll(HWND hwnd,unsigned int size)
{
  if(hwnd!=NULL)
  {
    MEScroll *ccp = GetMEScroll(hwnd);
    ccp->docLength=size;
    //is it disabled?
    if(ccp->docLength>ccp->screenSize)
    {
        ccp->disabled=FALSE;
        ShowWindow(hwnd,SW_SHOW);
        InvalidateRect(hwnd,NULL,FALSE);        
        UpdateWindow(hwnd); 
    }
    else 
    {
        ccp->disabled=TRUE;
        ShowWindow(hwnd,SW_HIDE);        
    }
                                                  
  }          
}





/**************************
*SET the THE SIZE OF THE VISIBLE PART OF THE DOCUMENT, in document units

return: void
see mescrollbar.h  or MEControls.h
for an explanation of the paramaters
***************************/
void setscreensizeMESCroll(HWND hwnd, int size)
{
  if(hwnd!=NULL)
  {
    MEScroll *ccp = GetMEScroll(hwnd);
    ccp->screenSize=size;
    if(ccp->docLength>ccp->screenSize)
    {
        ccp->disabled=FALSE;
        ShowWindow(hwnd,SW_SHOW);
        InvalidateRect(hwnd,NULL,FALSE);        
        UpdateWindow(hwnd); 
    }
    else 
    {
        ccp->disabled=TRUE;
        ShowWindow(hwnd,SW_HIDE);        
    }                                             
  }     
}


