/* ***********************************************************************************************
*metooltips.c													
*(c)Eduardo Rodriguez 2005																
*																					
*porpuse:   This creates a Custom tooltip for win32, intended to be schematic enough so as to
*           provide the basic functionality and allow for easy changes in the sourcecode, thus
*           not having to rely on 3rd party libraries and keeping the code tidy and small.
*           They don't do anything without the tooltip manager
*history:	16/11/2005 created															
*********************************************************************************************** */


#include <windows.h>
#include "MEControls.h"





/*************************
*DATA MEMBERS
**************************/

//info for every instance
typedef struct _METoolTip
{
    int width;    
    int height; 
    int posX;
    int posY;
    char *str;
}METoolTip;



const char g_METoolTips[]="metooltips";
extern HDC memdc;
extern HBITMAP bm;



//////////////////////////////////////////////////////////////////////////////////////
/********************************METHODS ("PRIVATE")********************************/
/////////////////////////////////////////////////////////////////////////////////////
//set and retrieve window info

static METoolTip*  GetMEToolTip(HWND hwnd)
{
    return (METoolTip *)GetWindowLong(hwnd, 0);
}



static void SetMEToolTip(HWND hwnd, METoolTip *ccp)
{
    SetWindowLong(hwnd, 0, (LONG)ccp);
}





/*************************
*PAINT
**************************/
//very simple painting proc, to be custumized. It does some
//calculations one shouldn't mess with, though


static void METoolTipPaint(HWND hwnd, METoolTip *ccp)
{
    //make sure it has been initiliazed 
    if(ccp!=NULL)
    {
       PAINTSTRUCT Ps;
       HDC hDC = BeginPaint(hwnd, &Ps);
       HDC MemDC=CreateCompatibleDC(hDC);
       HFONT hf = CreateFont(15, 0, 0, 0,0, FALSE, 0, 0, 0, 0, 0, 0, 0, "Times New Roman");
       COLORREF c1,c2,c3;
       HBRUSH bk1;
       HPEN p1;
       int kk=0,jj=0,height=5;
       char buffer[1000];
       memset(buffer,'\0',1000);
       SelectObject(memdc,hf);
       SelectObject(MemDC,bm);
       
       c1=GetPixel(MemDC,18,0);
       c2=GetPixel(MemDC,17,0);
       c3=GetPixel(MemDC,19,0);                
 
       bk1=CreateSolidBrush(c1);
       p1=CreatePen(PS_SOLID, 1, c2);
       SelectObject(memdc,bk1);
       SelectObject(memdc,p1);
       
       Rectangle(memdc,0,0,ccp->width,ccp->height);  
 
       SetTextColor(memdc, c3); 
       SetBkColor(memdc, c1); 
       
       while(ccp->str[kk]!='\0')
       {
           if(ccp->str[kk]=='\n')
           {
               TextOut(memdc,5,height,buffer,strlen(buffer));  
               height+=16;   
               memset(buffer,'\0',1000);
               jj=0;         
           }
           else
           {
               buffer[jj]=ccp->str[kk];
               jj++; 
           }             
           kk++;                   
       }

       //double buffer
       BitBlt(hDC, 0, 0,ccp->width,ccp->height , memdc, 0, 0, SRCCOPY);

       
       DeleteObject(hf);
       DeleteDC(MemDC);
       DeleteDC(hDC);
       EndPaint(hwnd, &Ps);
       

     
     
     
                 
    }     
}













/******************************
*CALLBACK PROCEDURE
*******************************/

LRESULT CALLBACK metooltipproc(HWND hwnd, UINT msg, WPARAM wParam, LPARAM lParam)
{
    METoolTip *ccp = GetMEToolTip(hwnd);

    switch(msg)
    {
     
      
         
        
         //-------------------------WM_NCDESTROY------------------------------------
         case WM_NCDESTROY:
             free(ccp);
         break;
         
         
         //---------------------------WM_PAINT-------------------------------------
         case WM_PAINT:
             METoolTipPaint(hwnd,ccp);
         break;
         
      case WM_ERASEBKGND:

       return 1;

         
         //blabla    
         default:
         break;
    }

    return DefWindowProc(hwnd, msg, wParam, lParam);
}






///////////////////////////////////////////////////////////////////////////////////////
/******************************METHODS ("PUBLIC")************************************/
//////////////////////////////////////////////////////////////////////////////////////

/*****************************
*REGISTER THE CLASS

return void;
HWND parent: any valid handle;
Should be initialized by Global
********************************/
void InitMEToolTip(HWND parent)
{

    WNDCLASSEX wc;

    wc.cbSize        = sizeof(WNDCLASSEX);
    wc.style         =  0;
    wc.lpfnWndProc   = metooltipproc;
    wc.cbClsExtra    = 0;
    wc.cbWndExtra    = sizeof( METoolTip * );
    wc.hInstance     = GetModuleHandle(0);
    wc.hIcon         = LoadIcon(NULL, IDI_APPLICATION);
    wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW+1);
    wc.lpszMenuName  = NULL;
    wc.lpszClassName = g_METoolTips;
    wc.hIconSm       = LoadIcon(NULL, IDI_APPLICATION);

    RegisterClassEx(&wc); 
}



/*****************************
*CREATE THE WINDOW

return handle to the newly created list;
See metooltips.h or MEControls.h for and explanation
of the parameters.
********************************/
HWND CreateMEToolTip(HWND parent,int left,int top, char* string)
{
 
   //calculate size
   SIZE size;
   int xx=GetSystemMetrics(SM_CXSCREEN);
   int yy=GetSystemMetrics(SM_CYSCREEN);
   int kk=0,jj=0,height=0,width=0;
   char buffer[1000];
   memset(buffer,'\0',1000);
   PAINTSTRUCT Ps;
   HDC hDC = BeginPaint(parent, &Ps);
   HDC hdc=CreateCompatibleDC(hDC);
   HFONT hf = CreateFont(15, 0, 0, 0,0, FALSE, 0, 0, 0, 0, 0, 0, 0, "Times New Roman");
   SelectObject(hdc,hf);
   while(string[kk]!='\0')
   {
       if(string[kk]=='\n')
       {
           height+=16; 
           GetTextExtentPoint32(hdc,buffer,strlen(buffer),&size); 
           if(size.cx>width)width=size.cx;
           memset(buffer,'\0',1000);
           jj=0;         
       }
       else
       {
           buffer[jj]=string[kk];
           jj++; 
       }             
       kk++;                   
   }
   
   DeleteObject(hf);
   DeleteDC(hdc);
   DeleteDC(hDC);
   EndPaint(parent, &Ps);
   if(left+width+10>xx)left=xx-(width+10);
   if(top+height+10>yy)top=top-(height+40);
   //create it
   HWND thisone=CreateWindowEx(0,g_METoolTips, "",WS_POPUP,left, top, width+10, height+10,
       parent, NULL, GetModuleHandle(0), NULL);  
   ShowWindow (thisone,  SW_SHOWNOACTIVATE);  
   METoolTip *ccp = GetMEToolTip(thisone);
   ccp = malloc( sizeof(METoolTip) );
   if(ccp == NULL) return NULL;

   // Initialize the CustCtrl structure. 
   ccp->width=width+10;
   ccp->height=height+10;
   ccp->str=string;
   SetMEToolTip(thisone,ccp);

 
return thisone;
  
     
}


